--[[
    RealGPS

    Shows the current GPS position of the player on a plane in the vehicle.

	@author: 		BayernGamers
	@date: 			11.03.2025
	@version:		2.0

	History:		v1.0 @15.07.2023 - initial implementation in FS 22
                    ------------------------------------------------------------------------------------------------------
                    v1.1 @16.07.2023 - update for Multi-PDA Support
                    ------------------------------------------------------------------------------------------------------
                    v1.2 @17.07.2023 - added glowShader to PDA
                    ------------------------------------------------------------------------------------------------------
                    v2.0 @11.03.2025 - convert and re-write for FS25
                    ------------------------------------------------------------------------------------------------------
	
	License:        Terms:
                        Usage:
                            Feel free to use this work as-is as long as you adhere to the following terms:
						Attribution:
							You must give appropriate credit to the original author when using this work.
						No Derivatives:
							You may not alter, transform, or build upon this work in any way.
						Usage: 
							The work may be used for personal and commercial purposes, provided it is not modified or adapted.
						Additional Clause:
							This script may not be converted, adapted, or incorporated into any other game versions or platforms except by GIANTS Software.
]]
source(Utils.getFilename("scripts/utils/LoggingUtil.lua", g_currentModDirectory))
source(Utils.getFilename("scripts/events/SetupModeEvent.lua", g_currentModDirectory))
source(Utils.getFilename("scripts/events/RotationModeEvent.lua", g_currentModDirectory))
source(Utils.getFilename("scripts/events/MapZoomEvent.lua", g_currentModDirectory))

local log = LoggingUtil.new(true, LoggingUtil.DEBUG_LEVELS.HIGH, "RealGPS.lua")

RealGPS = {}
RealGPS.MOD_DIR = g_currentModDirectory
RealGPS.MOD_NAME = g_currentModName
RealGPS.MOD_SETTINGS_DIR = g_currentModSettingsDirectory
RealGPS.BASE_PATH = "vehicle.realGPS"
RealGPS.CONTROLS_BASE_PATH = RealGPS.BASE_PATH .. ".controls"
RealGPS.MAP_BASE_PATH = RealGPS.BASE_PATH .. ".map"
RealGPS.PROJECTION_BASE_PATH = RealGPS.MAP_BASE_PATH .. ".projection"

RealGPS.defaultVehicles = {}
RealGPS.DEFAULT_VEHICLE_BASE_PATH = "defaultVehicles.vehicle(?)"
RealGPS.DEFAULT_VEHICLE_XML = RealGPS.MOD_DIR .. "xml/defaultVehicles.xml"
RealGPS.DEFAULT_VEHICLE_SCHEMA = XMLSchema.new("realGPSDefaultVehicles")
createFolder(RealGPS.MOD_SETTINGS_DIR)

function RealGPS.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Enterable, specializations) and not SpecializationUtil.hasSpecialization(Locomotive, specializations)
end

function RealGPS.initSpecialization()
    local schema = Vehicle.xmlSchema
    schema:setXMLSpecializationType("RealGPS")

    schema:register(XMLValueType.BOOL, RealGPS.BASE_PATH .. "#isMotorActiveDependent", "Disable PDA when motor is off", false)
    schema:register(XMLValueType.BOOL, RealGPS.BASE_PATH .. "#isIgnitionDependent", "Disable PDA when the ignition is off", false)

    schema:register(XMLValueType.BOOL, RealGPS.CONTROLS_BASE_PATH .. "#allow", "Allow the player to control the PDA", true)
    schema:register(XMLValueType.BOOL, RealGPS.CONTROLS_BASE_PATH .. "#allowRotation", "Allow the player to rotate the PDA", true)
    schema:register(XMLValueType.BOOL, RealGPS.CONTROLS_BASE_PATH .. "#allowZoom", "Allow the player to zoom the PDA", true)

    schema:register(XMLValueType.NODE_INDEX, RealGPS.MAP_BASE_PATH .. "#linkNode", "Node to load the PDA into")
    schema:register(XMLValueType.NODE_INDEX, RealGPS.MAP_BASE_PATH .. "#hideNode", "Node to hide the original Dashboards")
    schema:register(XMLValueType.FLOAT, RealGPS.MAP_BASE_PATH .. "#intensity", "Intensity of the glow shader", 1.0)
    schema:register(XMLValueType.FLOAT, RealGPS.MAP_BASE_PATH .. "#idleValue", "Value to set the glow shader to when the PDA is not active", -1)
    schema:register(XMLValueType.BOOL, RealGPS.MAP_BASE_PATH .. "#toggleVisibility", "Toggle the visibility of the PDA instead of the idleValue", false)
    schema:register(XMLValueType.FLOAT, RealGPS.MAP_BASE_PATH .. "#stateDelay", "Delay before the PDA is shown", 0.0)

    schema:register(XMLValueType.STRING, RealGPS.PROJECTION_BASE_PATH .. "#mode", "Projection mode of the PDA", "PLAYER")
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#playerSize", "Size of the player marker", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#playerSizeMax", "Max size of the player marker", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#playerSizeMin", "Min size of the player marker", 0.001)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoom", "Zoom of the background", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoomMax", "Max zoom of the background", 20)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoomMin", "Min zoom of the background", 1.35)
    schema:register(XMLValueType.FLOAT, RealGPS.PROJECTION_BASE_PATH .. "#aspectRatio", "Aspect ratio of the PDA", 1)

    schema:setXMLSpecializationType()

    schema = RealGPS.DEFAULT_VEHICLE_SCHEMA
    schema:setXMLSpecializationType("RealGPSDefaultVehicle")

    schema:register(XMLValueType.STRING, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. "#filename", "Filename of the vehicle", nil)
    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. "#isMotorActiveDependent", "Disable PDA when motor is off", false)
    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. "#isIgnitionDependent", "Disable PDA when the ignition is off", false)

    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".controls#allow", "Allow the player to control the PDA", true)
    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".controls#allowRotation", "Allow the player to rotate the PDA", true)
    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".controls#allowZoom", "Allow the player to zoom the PDA", true)

    schema:register(XMLValueType.STRING, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#rootNode", "Node to load the PDA into")
    schema:register(XMLValueType.STRING, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#hideNode", "Node to hide the original Dashboards")
    schema:register(XMLValueType.VECTOR_TRANS, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#translation", "Translation of the PDA")
    schema:register(XMLValueType.VECTOR_ROT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#rotation", "Rotation of the PDA")
    schema:register(XMLValueType.VECTOR_SCALE, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#scale", "Scale of the PDA")

    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#intensity", "Intensity of the glow shader", 1.0)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#idleValue", "Value to set the glow shader to when the PDA is not active", -1)
    schema:register(XMLValueType.BOOL, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#toggleVisibility", "Toggle the visibility of the PDA instead of the idleValue", false)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map#stateDelay", "Delay before the PDA is shown", 0.0)

    schema:register(XMLValueType.STRING, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#mode", "Projection mode of the PDA", "PLAYER")
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#playerSize", "Size of the player marker", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#playerSizeMax", "Max size of the player marker", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#playerSizeMin", "Min size of the player marker", 0.001)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#backgroundZoom", "Zoom of the background", 1)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#backgroundZoomMax", "Max zoom of the background", 20)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#backgroundZoomMin", "Min zoom of the background", 1.35)
    schema:register(XMLValueType.FLOAT, RealGPS.DEFAULT_VEHICLE_BASE_PATH .. ".map.projection#aspectRatio", "Aspect ratio of the PDA", 1)

    schema:setXMLSpecializationType()
end

function RealGPS.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", RealGPS)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", RealGPS)
    SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", RealGPS)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdate", RealGPS)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", RealGPS)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", RealGPS)

	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", RealGPS)
end

function RealGPS.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "loadGPSMap", RealGPS.loadGPSMap)
    SpecializationUtil.registerFunction(vehicleType, "setMapZoom", RealGPS.setMapZoom)
    SpecializationUtil.registerFunction(vehicleType, "setPlayerSize", RealGPS.setPlayerSize)
    SpecializationUtil.registerFunction(vehicleType, "setRotateMap", RealGPS.setRotateMap)
    SpecializationUtil.registerFunction(vehicleType, "setMotorDependentVisibility", RealGPS.setMotorDependentVisibility)
    SpecializationUtil.registerFunction(vehicleType, "calculateMapAndPlayerPosition", RealGPS.calculateMapAndPlayerPosition)
    SpecializationUtil.registerFunction(vehicleType, "calculateRootWorldYAxisRotation", RealGPS.calculateRootWorldYAxisRotation)
    SpecializationUtil.registerFunction(vehicleType, "mapRange", RealGPS.mapRange)
    SpecializationUtil.registerFunction(vehicleType, "getIsIgnitionActive", RealGPS.getIsIgnitionActive)
    SpecializationUtil.registerFunction(vehicleType, "loadDefaultVehicles", RealGPS.loadDefaultVehicles)
    SpecializationUtil.registerFunction(vehicleType, "resetDelayTimer", RealGPS.resetDelayTimer)
end

function RealGPS:saveToXMLFile(xmlFile, key, usedModNames)
    local spec = self.spec_realGPS

    if spec.hasRealGPSMap == true then
        local mapKey = key .. ".map"
        setXMLFloat(xmlFile.handle, mapKey .. "#mapZoom", spec.map.zoom)
        setXMLFloat(xmlFile.handle, mapKey .. "#playerSize", spec.projection.playerSize)
        setXMLBool(xmlFile.handle, mapKey .. "#rotateMap", spec.map.rotatingMap)
	end
end

function RealGPS:onReadStream(streamId, connection)
    local spec = self.spec_realGPS

	if spec.hasRealGPSMap == true then
        spec.isInSetupMode = streamReadBool(streamId)
        self:setMapZoom(streamReadFloat32(streamId))
        self:setPlayerSize(streamReadFloat32(streamId))
        self:setRotateMap(streamReadBool(streamId))
	end
end

function RealGPS:onWriteStream(streamId, connection)
    local spec = self.spec_realGPS

	if spec.hasRealGPSMap == true then
        streamWriteBool(streamId, spec.isInSetupMode)
        streamWriteFloat32(streamId, spec.map.zoom)
        streamWriteFloat32(streamId, spec.projection.playerSize)
        streamWriteBool(streamId, spec.map.rotatingMap)
	end
end

function RealGPS:onPreLoad(savegame)
    self.spec_realGPS = {}
    local spec = self.spec_realGPS
end

function RealGPS:onLoad(savegame)
    local spec = self.spec_realGPS

    spec.hasRealGPSMap = false

    self:loadDefaultVehicles()

    local dlcFilename = nil
    if self.xmlFile:getFilename():find("pdlc") ~= nil then
        local pattern = ".*pdlc"
        dlcFilename = "pdlc" .. self.xmlFile:getFilename():gsub(pattern, "")
    end

    if (self.xmlFile:getFilename() ~= nil and RealGPS.defaultVehicles[self.xmlFile:getFilename()] ~= nil) or (dlcFilename ~= nil and RealGPS.defaultVehicles[dlcFilename] ~= nil) then
        local defaultVehicle = nil

        if dlcFilename ~= nil then
            defaultVehicle = RealGPS.defaultVehicles[dlcFilename]
        else
            defaultVehicle = RealGPS.defaultVehicles[self.xmlFile:getFilename()]
        end

        spec.isIgnitionDependent = defaultVehicle.isIgnitionDependent
        spec.isMotorActiveDependent = defaultVehicle.isMotorActiveDependent

        if spec.isIgnitionDependent == true and spec.isMotorActiveDependent == true then
            spec.isMotorActiveDependent = false
        end

        spec.isLoading = true

        spec.controls = {}
        spec.controls.allow = defaultVehicle.controls.allow
        spec.controls.allowRotation = defaultVehicle.controls.allowRotation
        spec.controls.allowZoom = defaultVehicle.controls.allowZoom

        local linkNode = createTransformGroup("realGPSMap")
        local rootNode = I3DUtil.indexToObject(self.components, defaultVehicle.map.rootNode, self.i3dMappings)

        link(rootNode, linkNode)
        setTranslation(linkNode, unpack(defaultVehicle.map.translation))
        setRotation(linkNode, unpack(defaultVehicle.map.rotation))
        setScale(linkNode, unpack(defaultVehicle.map.scale))

        if defaultVehicle.map.hideNode ~= nil then
            local hideNode = I3DUtil.indexToObject(self.components, defaultVehicle.map.hideNode, self.i3dMappings)
            setVisibility(hideNode, false)
        end
    
        spec.map = {}
        spec.map.linkNode = linkNode
        spec.map.intensity = defaultVehicle.map.intensity
        spec.map.idleValue = defaultVehicle.map.idleValue
        spec.map.toggleVisibility = defaultVehicle.map.toggleVisibility
        spec.map.stateDelay = defaultVehicle.map.stateDelay
        self:resetDelayTimer()

        spec.actionEvents = {}
        spec.isInSetupMode = false

        spec.projection = {}
        spec.projection.mode = defaultVehicle.projection.mode
        spec.projection.playerSize = defaultVehicle.projection.playerSize
        spec.projection.playerSizeMax = defaultVehicle.projection.playerSizeMax
        spec.projection.playerSizeMin = defaultVehicle.projection.playerSizeMin
        spec.projection.backgroundZoom = defaultVehicle.projection.backgroundZoom
        spec.projection.backgroundZoomMax = defaultVehicle.projection.backgroundZoomMax
        spec.projection.backgroundZoomMin = defaultVehicle.projection.backgroundZoomMin
        spec.projection.aspectRatio = defaultVehicle.projection.aspectRatio

        if g_currentMission:getIsClient() then
            self:loadGPSMap()
        end

        spec.hasRealGPSMap = true
    end


    if self.xmlFile:hasProperty(RealGPS.BASE_PATH) and not spec.hasRealGPSMap then
        spec.isIgnitionDependent = self.xmlFile:getValue(RealGPS.BASE_PATH .. "#isIgnitionDependent", false)
        spec.isMotorActiveDependent = self.xmlFile:getValue(RealGPS.BASE_PATH .. "#isMotorActiveDependent", false)

        if spec.isIgnitionDependent == true and spec.isMotorActiveDependent == true then
            spec.isMotorActiveDependent = false
        end

        spec.isLoading = true

        spec.controls = {}
        spec.controls.allow = self.xmlFile:getValue(RealGPS.CONTROLS_BASE_PATH .. "#allow", true)
        spec.controls.allowRotation = self.xmlFile:getValue(RealGPS.CONTROLS_BASE_PATH .. "#allowRotation", true)
        spec.controls.allowZoom = self.xmlFile:getValue(RealGPS.CONTROLS_BASE_PATH .. "#allowZoom", true)

        spec.map = {}
        spec.map.linkNode = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#linkNode", nil, self.components, self.i3dMappings)
        spec.map.intensity = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#intensity", 1.0)
        spec.map.idleValue = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#idleValue", -1)
        spec.map.toggleVisibility = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#toggleVisibility", false)
        spec.map.stateDelay = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#stateDelay", 0.0)
        self:resetDelayTimer()

        local hideNode = self.xmlFile:getValue(RealGPS.MAP_BASE_PATH .. "#hideNode", nil, self.components, self.i3dMappings)
        if hideNode ~= nil then
            setVisibility(hideNode, false)
        end

        spec.actionEvents = {}
        spec.isInSetupMode = false

        spec.projection = {}
        spec.projection.mode = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#mode", "PLAYER")
        spec.projection.playerSize = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#playerSize", 1)
        spec.projection.playerSizeMax = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#playerSizeMax", 0.5)
        spec.projection.playerSizeMin = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#playerSizeMin", 0.001)
        spec.projection.backgroundZoom = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoom", 1)
        spec.projection.backgroundZoomMax = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoomMax", 1)
        spec.projection.backgroundZoomMin = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#backgroundZoomMin", 1)
        spec.projection.aspectRatio = self.xmlFile:getValue(RealGPS.PROJECTION_BASE_PATH .. "#aspectRatio", 1)

        if g_currentMission:getIsClient() then
            self:loadGPSMap()
        end

        spec.hasRealGPSMap = true
    end
end

function RealGPS:onPostLoad(savegame)
    local spec = self.spec_realGPS

    if spec.hasRealGPSMap == true then
        local spec = self.spec_realGPS
        local zoom = 2
        local playerSize = 0.03125
        local rotateMap = spec.projection.mode ~= "PLAYER"

        if savegame ~= nil then
            local key = savegame.key .. "." .. RealGPS.MOD_NAME .. ".realGPS.map"

            zoom       = Utils.getNoNil(getXMLFloat(savegame.xmlFile.handle, key .. "#mapZoom"),    zoom       )
            playerSize = Utils.getNoNil(getXMLFloat(savegame.xmlFile.handle, key.. "#playerSize"), playerSize )
            rotateMap  = Utils.getNoNil(getXMLBool(savegame.xmlFile.handle, key .. "#rotateMap"),   rotateMap  )
        end

        self:setMapZoom(zoom)
        self:setPlayerSize(playerSize)
        self:setRotateMap(rotateMap)

        if g_currentMission:getIsClient() then
            setShaderParameter(spec.map.shaderObject , "mapProps"    , 0.5, 0.5, spec.map.zoom, 0, false)
            local aspectX = 1
            local aspectY = 1
            if spec.projection.aspectRatio > 1 then
                aspectX = 1/spec.projection.aspectRatio
            else
                aspectY = spec.projection.aspectRatio
            end
            setShaderParameter(spec.map.shaderObject , "screenProps" , aspectX, aspectY, 0, 0, false)
        end
    end
end

function RealGPS:onUpdate(dt)
    local spec = self.spec_realGPS

    if g_currentMission:getIsClient() and self:getIsActive() and spec.hasRealGPSMap == true and self.rootNode ~= nil then
        self:setMotorDependentVisibility()

        local x, _, z = getWorldTranslation(self.rootNode)

        if spec.map.rotatingMap then
            local mapXval = self:mapRange(g_currentMission.hud.ingameMap.normalizedPlayerPosX * g_currentMission.hud.ingameMap.worldSizeX, 0, g_currentMission.hud.ingameMap.worldSizeX, 0.25, 0.75)
            local mapYval = self:mapRange(g_currentMission.hud.ingameMap.normalizedPlayerPosZ * g_currentMission.hud.ingameMap.worldSizeZ, 0, g_currentMission.hud.ingameMap.worldSizeX, 0.75, 0.25)

            setShaderParameter(spec.map.shaderObject , "mapProps", mapXval, mapYval, spec.map.zoom, math.pi - self:calculateRootWorldYAxisRotation(), false)
            setShaderParameter(spec.map.shaderObject , "playerProps", 0.5, 0.5, math.pi, spec.projection.playerSize, false)
        else
            local mapXval = self:mapRange(g_currentMission.hud.ingameMap.normalizedPlayerPosX * g_currentMission.hud.ingameMap.worldSizeX, 0, g_currentMission.hud.ingameMap.worldSizeX, 0.25, 0.75)
            local mapYval = self:mapRange(g_currentMission.hud.ingameMap.normalizedPlayerPosZ * g_currentMission.hud.ingameMap.worldSizeZ, 0, g_currentMission.hud.ingameMap.worldSizeX, 0.75, 0.25)

            --print("mapXVal: " .. mapXval)
            --print("mapYVal " .. mapYval)
            --print("Zoom: " .. self.LRM.map.zoom)

            local mapX, playerX = self:calculateMapAndPlayerPosition(spec.map.mapHighLimit, spec.map.mapLowLimit, mapXval, spec.map.zoom)
            local mapY, playerY = self:calculateMapAndPlayerPosition(spec.map.mapHighLimit, spec.map.mapLowLimit, mapYval, spec.map.zoom)

            local rotation = self:calculateRootWorldYAxisRotation()

            --print("Player X:" .. playerX .. " PlayerY: " .. playerY .. " MapX: " .. mapX .. " MapY: " .. mapY)
            
            setShaderParameter(spec.map.shaderObject , "playerProps", playerX, playerY, rotation, spec.projection.playerSize, false)
            setShaderParameter(spec.map.shaderObject , "mapProps", mapX, mapY, spec.map.zoom, 0, false )	
        end
    end
end

function RealGPS:onUpdateTick(dt)
    local spec = self.spec_realGPS

    if g_currentMission:getIsClient() and self:getIsActive() and spec.hasRealGPSMap == true and self.rootNode ~= nil then
        if spec.map.startDelayTimer == true then
            spec.map.delayTimer = spec.map.delayTimer + dt

            if spec.map.delayTimer >= spec.map.stateDelay then
                spec.map.startDelayTimer = false
                spec.map.delayTimer = 0
                spec.map.delayTimerFinished = true
            end
        end
    end
end

function RealGPS:resetDelayTimer()
    local spec = self.spec_realGPS

    spec.map.startDelayTimer = false
    spec.map.delayTimer = 0
    spec.map.delayTimerFinished = false
end

function RealGPS:onDelete()
    local spec = self.spec_realGPS

    if g_currentMission:getIsClient() then
        if spec ~= nil and spec.map ~= nil and spec.map.sharedLoadRequestId ~= nil then
            g_i3DManager:releaseSharedI3DFile(spec.map.sharedLoadRequestId)
            spec.map.mapObject = nil
            spec.map.sharedLoadRequestId = nil
        end
    end
end

function RealGPS:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    local spec = self.spec_realGPS

    if g_currentMission:getIsClient() and spec.hasRealGPSMap == true then
        self:clearActionEventsTable(spec.actionEvents)

        if isActiveForInputIgnoreSelection then
            local _, actionEventId = self:addActionEvent(spec.actionEvents, "RGPS_SETUP_MODE", self, RealGPS.actionEvent_toggleSetupMode, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL)
            spec.actionEventId = actionEventId

            local _, actionEventId1 = self:addActionEvent(spec.actionEvents, "RGPS_TOGGLE_ROTATE_MODE", self, RealGPS.actionEvent_toggleRotateMode, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId1, GS_PRIO_NORMAL)
            spec.actionEventId1 = actionEventId1
			
			local _, actionEventId2 = self:addActionEvent(spec.actionEvents, "RGPS_AXIS_ZOOM", self, RealGPS.actionEvent_zoom, false, true, true, true, nil, nil, true)
			spec.actionEventId2 = actionEventId2
			g_inputBinding:setActionEventTextPriority(actionEventId2, GS_PRIO_NORMAL)

			RealGPS.updateActionEvents(self)
        end
    end
end

function RealGPS:updateActionEvents()
    local spec = self.spec_realGPS

	g_inputBinding:setActionEventText(spec.actionEventId, spec.isInSetupMode and g_i18n:getText("input_RGPS_toggleSetupModeDisable") or g_i18n:getText("input_RGPS_toggleSetupModeEnable"))
    g_inputBinding:setActionEventText(spec.actionEventId1, spec.map.rotatingMap and g_i18n:getText("input_RGPS_toggleRotateModeDisable") or g_i18n:getText("input_RGPS_toggleRotateModeEnable"))
    g_inputBinding:setActionEventText(spec.actionEventId2, g_i18n:getText("input_RGPS_zoom"))

    g_inputBinding:setActionEventActive(spec.actionEventId1, spec.isInSetupMode)
	g_inputBinding:setActionEventActive(spec.actionEventId2, spec.isInSetupMode)
end

function RealGPS:loadGPSMap()
    local spec = self.spec_realGPS

    if spec.map.linkNode ~= nil then
        local node, sharedLoadRequestId, failedReason = g_i3DManager:loadSharedI3DFile(RealGPS.MOD_SETTINGS_DIR .. "pdaMap/pdaMap.i3d", false, false)

        if failedReason ~= 0 then
            log:printError("Failed to load the PDA map. Error Id: " .. failedReason)
        end
    
        spec.map.mapObject = node
        spec.map.sharedLoadRequestId = sharedLoadRequestId

        link(spec.map.linkNode, spec.map.mapObject) 
        local mapObject = getChildAt(spec.map.linkNode, 0)
        spec.map.shaderObject = getChildAt(mapObject, 0)
    end
end

function RealGPS:calculateRootWorldYAxisRotation()
    local object = self.rootNode
	local alpha, beta, gamma = getRotation(object)
	local yAxis = math.cos(gamma)*math.sin(beta)*math.cos(alpha) + math.sin(gamma)*math.sin(alpha)
	local xAxis = math.cos(beta)*math.cos(alpha)
	return(math.atan2(yAxis, xAxis))
end

function RealGPS:mapRange(value, fromMin, fromMax, toMin, toMax)
	local fromRange = fromMax - fromMin
	local toRange = toMax - toMin
	local scaledValue = (value - fromMin) / fromRange
	local mappedValue = toMin + (scaledValue * toRange)
	return mappedValue
end

function RealGPS:setMotorDependentVisibility()
    local spec = self.spec_realGPS

    if g_currentMission:getIsClient() then

        if (self.propertyState == VehiclePropertyState.SHOP_CONFIG or not self.getIsEntered) and spec.map.shaderObject ~= nil then
            if spec.map.toggleVisibility then
                setVisibility(spec.map.shaderObject, false)
            else
                setShaderParameter(spec.map.shaderObject, "lightControl", spec.map.idleValue, 0, 0, 0, false)
            end
        else
            local shaderParameter = "lightControl"
            local x, y, z, w = getShaderParameter(spec.map.shaderObject, shaderParameter)

            if spec.isMotorActiveDependent == true then
                if self:getIsMotorStarted() and spec.map.stateDelay > 0 and not spec.map.delayTimerFinished and not spec.map.startDelayTimer then
                    spec.map.startDelayTimer = true
                end

                if spec.map.toggleVisibility then
                    if self:getIsMotorStarted() and x ~= spec.map.intensity and (spec.map.stateDelay == 0 or spec.map.delayTimerFinished) then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.intensity, 0, 0, 0, false)
                        setVisibility(spec.map.shaderObject, true)
                    elseif not self:getIsMotorStarted() and x ~= spec.map.idleValue then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.idleValue, 0, 0, 0, false)
                        setVisibility(spec.map.shaderObject, false)
                        self:resetDelayTimer()
                    end
                else
                    if self:getIsMotorStarted() and x ~= spec.map.intensity and (spec.map.stateDelay == 0 or spec.map.delayTimerFinished) then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.intensity, 0, 0, 0, false)
                    elseif not self:getIsMotorStarted() and x ~= spec.map.idleValue then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.idleValue, 0, 0, 0, false)
                        self:resetDelayTimer()
                    end
                end
            elseif spec.isIgnitionDependent == true then
                if self:getIsIgnitionActive() and spec.map.stateDelay > 0 and not spec.map.delayTimerFinished and not spec.map.startDelayTimer then
                    spec.map.startDelayTimer = true
                end

                if spec.map.toggleVisibility then
                    if self:getIsIgnitionActive() and x ~= spec.map.intensity and (spec.map.stateDelay == 0 or spec.map.delayTimerFinished) then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.intensity, 0, 0, 0, false)
                        setVisibility(spec.map.shaderObject, true)
                    elseif (not self:getIsIgnitionActive() and x ~= spec.map.idleValue) or spec.isLoading then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.idleValue, 0, 0, 0, false)
                        setVisibility(spec.map.shaderObject, false)
                        self:resetDelayTimer()
                        spec.isLoading = false
                    end
                else
                    if self:getIsIgnitionActive() and x ~= spec.map.intensity and (spec.map.stateDelay == 0 or spec.map.delayTimerFinished) then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.intensity, 0, 0, 0, false)
                    elseif (not self:getIsIgnitionActive() and x ~= spec.map.idleValue) or spec.isLoading then
                        setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.idleValue, 0, 0, 0, false)
                        self:resetDelayTimer()
                        spec.isLoading = false
                    end
                end
            elseif x ~= spec.map.intensity then
                setShaderParameter(spec.map.shaderObject, shaderParameter, spec.map.intensity, 0, 0, 0, false)
            end
        end
    end
end

function RealGPS:calculateMapAndPlayerPosition(highLimit, lowLimit, value, zoom)
    local spec = self.spec_realGPS

	if value <= lowLimit then
		return lowLimit, (value * zoom - 0.5 * zoom / 2)
	elseif value >= highLimit then
		return highLimit, self:mapRange(value, spec.map.mapHighLimit, 0.75, 0.5, 1)
	else
		return value, 0.5
	end
end

function RealGPS:setMapZoom(value)
    local spec = self.spec_realGPS
    spec.map.zoom = value
    
    if g_currentMission:getIsClient() then
        spec.map.mapLowLimit  = 0.5 / spec.map.zoom + 0.25
        spec.map.mapHighLimit = 1 - spec.map.mapLowLimit
    end
end

function RealGPS:setPlayerSize(value)
    local spec = self.spec_realGPS
    spec.projection.playerSize = value
    
    if g_currentMission:getIsClient() then
        if spec.map.rotatingMap then
            setShaderParameter(spec.map.shaderObject , "playerProps", 0.5, 0.5, math.pi, spec.projection.playerSize, false)
        end
    end
end

function RealGPS:setRotateMap(value)
    local spec = self.spec_realGPS
    spec.map.rotatingMap = value
    
    if g_currentMission:getIsClient() then
        if spec.map.rotatingMap then
            setShaderParameter(spec.map.shaderObject , "playerProps", 0.5, 0.5, math.pi, spec.projection.playerSize, false)
        end
    end
end

function RealGPS:actionEvent_toggleSetupMode(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_realGPS
	spec.isInSetupMode = not spec.isInSetupMode
	SetupModeEvent.sendEvent(self, spec.isInSetupMode, false)
	RealGPS.updateActionEvents(self)
end

function RealGPS:actionEvent_toggleRotateMode(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_realGPS
	if spec.isInSetupMode then
		if spec.controls.allowRotation then
            self:setRotateMap(not spec.map.rotatingMap)
            RealGPS.updateActionEvents(self)
            RotationModeEvent.sendEvent(self, spec.map.rotatingMap, false)
		end
	end
end

function RealGPS:actionEvent_zoom(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_realGPS
    if spec.isInSetupMode then
        local newZoom = spec.map.zoom + inputValue * 0.1
        local allowPlayerSizing = true

        if newZoom < spec.projection.backgroundZoomMin then
            newZoom = spec.projection.backgroundZoomMin
            allowPlayerSizing = false
        elseif newZoom > spec.projection.backgroundZoomMax then
            newZoom = spec.projection.backgroundZoomMax
            allowPlayerSizing = false
        end

        if allowPlayerSizing then
            local newSize = spec.projection.playerSize + ((inputValue * 0.01) / (spec.projection.backgroundZoomMax - spec.projection.backgroundZoomMin) / 2)

            if newSize < spec.projection.playerSizeMin then
                newSize = spec.projection.playerSizeMin
            elseif newSize > spec.projection.playerSizeMax then
                newSize = spec.projection.playerSizeMax
            end
            self:setPlayerSize(newSize)
        end

        self:setMapZoom(newZoom)
        MapZoomEvent.sendEvent(self, spec.map.zoom, false)
    end
end

function RealGPS:getIsIgnitionActive()
    if g_ignitionLockManager:getIsAvailable() then
        local ignitionState = g_ignitionLockManager:getState()
        if ignitionState == IgnitionLockState.OFF then
            return false
        elseif ignitionState == IgnitionLockState.IGNITION or ignitionState == IgnitionLockState.START then
            return true
        end
    else
        local motorState = self:getMotorState()
        if motorState == MotorState.ON or motorState == MotorState.STARTING or motorState == MotorState.IGNITION then
            return true
        else
            return false
        end
    end
end

function RealGPS:loadDefaultVehicles()
    RealGPS.defaultVehicles = {}
    local defaultVehiclesXML = XMLFile.loadIfExists("realGPSDefaultVehicles", RealGPS.DEFAULT_VEHICLE_XML, RealGPS.DEFAULT_VEHICLE_SCHEMA)

    if defaultVehiclesXML ~= nil then
        defaultVehiclesXML:iterate(RealGPS.DEFAULT_VEHICLE_BASE_PATH:sub(1, -4), function (index, key)
            local filename = defaultVehiclesXML:getString(key .. "#filename")
            filename = Utils.getFilename(filename)

            if RealGPS.defaultVehicles[filename] == nil then
                RealGPS.defaultVehicles[filename] = {}
                local defaultVehicle = RealGPS.defaultVehicles[filename]

                defaultVehicle.isMotorActiveDependent = defaultVehiclesXML:getValue(key .. "#isMotorActiveDependent", false)
                defaultVehicle.isIgnitionDependent = defaultVehiclesXML:getValue(key .. "#isIgnitionDependent", false)

                defaultVehicle.controls = {}
                defaultVehicle.controls.allow = defaultVehiclesXML:getValue(key .. ".controls#allow", true)
                defaultVehicle.controls.allowRotation = defaultVehiclesXML:getValue(key .. ".controls#allowRotation", true)
                defaultVehicle.controls.allowZoom = defaultVehiclesXML:getValue(key .. ".controls#allowZoom", true)

                defaultVehicle.map = {}
                
                defaultVehicle.map.rootNode = defaultVehiclesXML:getValue(key .. ".map#rootNode", nil)
                defaultVehicle.map.hideNode = defaultVehiclesXML:getValue(key .. ".map#hideNode", nil)
                defaultVehicle.map.translation = defaultVehiclesXML:getValue(key .. ".map#translation", "0 0 0", true)
                defaultVehicle.map.rotation = defaultVehiclesXML:getValue(key .. ".map#rotation", "0 0 0", true)
                defaultVehicle.map.scale = defaultVehiclesXML:getValue(key .. ".map#scale", "1 1 1", true)
                defaultVehicle.map.intensity = defaultVehiclesXML:getValue(key .. ".map#intensity", 1.0)
                defaultVehicle.map.idleValue = defaultVehiclesXML:getValue(key .. ".map#idleValue", -1)
                defaultVehicle.map.toggleVisibility = defaultVehiclesXML:getValue(key .. ".map#toggleVisibility", false)
                defaultVehicle.map.stateDelay = defaultVehiclesXML:getValue(key .. ".map#stateDelay", 0.0)

                defaultVehicle.projection = {}
                defaultVehicle.projection.mode = defaultVehiclesXML:getValue(key .. ".map.projection#mode", "PLAYER")
                defaultVehicle.projection.playerSize = defaultVehiclesXML:getValue(key .. ".map.projection#playerSize", 1)
                defaultVehicle.projection.playerSizeMax = defaultVehiclesXML:getValue(key .. ".map.projection#playerSizeMax", 0.5)
                defaultVehicle.projection.playerSizeMin = defaultVehiclesXML:getValue(key .. ".map.projection#playerSizeMin", 0.001)
                defaultVehicle.projection.backgroundZoom = defaultVehiclesXML:getValue(key .. ".map.projection#backgroundZoom", 1)
                defaultVehicle.projection.backgroundZoomMax = defaultVehiclesXML:getValue(key .. ".map.projection#backgroundZoomMax", 1)
                defaultVehicle.projection.backgroundZoomMin = defaultVehiclesXML:getValue(key .. ".map.projection#backgroundZoomMin", 1)
                defaultVehicle.projection.aspectRatio = defaultVehiclesXML:getValue(key .. ".map.projection#aspectRatio", 1)
            end
        end)
    end
end